;; Generated package description from ssh-deploy.el  -*- no-byte-compile: t -*-
(define-package "ssh-deploy" "3.1.16.0.20230702.92809" "Deployment via Tramp, global or per directory." '((emacs "25")) :commit "95fb076c9b657c5f1bfad3ee5bf1f8691c50d428" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://github.com/cjohansson/emacs-ssh-deploy")
